/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Abacus Template Editor Dialog.
 *
 * The Initial Developer of the Original Code is 
 * Alexander J. Vincent <ajvincent@juno.com>.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

function testTemplate() {
  var testDocString = "";
  testDocString += '<?xml-stylesheet type="text/css" href="chrome://abacus/content/tools/integerControl.css"?>\n';
  testDocString += '<?xml-stylesheet type="text/css" href="chrome://abacus/content/tools/mEdit.css"?>\n';
  testDocString += '<!DOCTYPE templates [\n';
  testDocString += '<!ENTITY % namespaces.dtd SYSTEM "chrome://abacus/content/tools/namespaces.dtd"> %namespaces.dtd;';
  testDocString += '\n]>\n';
  testDocString += '<templates xmlns="'+ MEDIT_NS + '#templates">';
  testDocString += prefixText.value + mainText.value + postfixText.value;
  testDocString += '</templates>';
  var testDoc = parser.parseFromString(testDocString, "application/xml");
  if (!parser.checkRootElement(testDoc, "templates")) {
    mainText.setAttribute("invalid", true);
    return false;
  }
  mainText.removeAttribute("invalid");
  removeWhitespaceByNamespace(testDoc, MEDIT_NS + "#templates");
  removeWhitespaceByNamespace(testDoc, MEDIT_NS);
  removeWhitespaceByNamespace(testDoc, MATH_NS);
  return testDoc;
}

function showTemplate() {
  var testDoc = testTemplate();
  if (!testDoc) {
    return false;
  }
  editor.showControls(testDoc.documentElement.firstChild.childNodes.item(1));
  editor.insertControlsNode = testDoc.documentElement;
  editor.selectedTemplate = testDoc.documentElement.firstChild.firstChild.cloneNode(true);
  window.setTimeout(updateEditorTemplate, 0); // This is done to allow the mEdit:control bindings to take effect first.
  return true;
}

const editor = {
  controlsBox:  null,
  templatesBox: null,
  insertControlsNode: null,
  hasInitialized: false,
  nodesMarkedForDeath: [],
  selectedTemplate: null,
  commandsArray: null,
  acceptButton: null,

  init: function() {
    this.controlsBox    = document.getElementById("controlsBox");
    this.templatesBox   = document.getElementById("templatesBox");
    this.commandsArray  = document.getElementById("abacusTemplateCommands").childNodes;
    this.acceptButton   = document.documentElement.getButton('accept');
    this.hasInitialized = true;
    this.updateCommandsDisabled(null);
  },

  showControls: function(aControlList) {
    // show mEdit:control elements
    while (this.controlsBox.firstChild) {
      this.controlsBox.removeChild(this.controlsBox.firstChild);
    }

    if (aControlList) {
      assertAboutInput(aControlList.localName == "controlList", "Invalid control list in editor.showControls!", true);
      while (aControlList.firstChild) {
        this.controlsBox.insertBefore(aControlList.firstChild, this.controlsBox.lastChild);
      }
      aControlList.parentNode.removeChild(aControlList);
    }
  },

  updateTemplate: function() {
    var i;

    this.templatesBox.controlList = [];
    for (i = 0; i < this.controlsBox.childNodes.length - 1; i++) {
      var aControlNode = this.controlsBox.childNodes[i];
      this.templatesBox.controlList.push({name: aControlNode.controlName, value: aControlNode.value});
    }

    while (this.templatesBox.firstChild) {
      this.templatesBox.removeChild(this.templatesBox.firstChild);
    }

    this.templatesBox.appendChild(this.selectedTemplate.cloneNode(true));
    return true;
  },

  markNodeForDeath: function(aNode) {
    /* This function sets us up to wipe out a node after all the events have taken place,
     * forcing a repaint of the XUL canvas in that node's location.
     * Sad experience and many hours of late-night coding have shown this 
     * to be a necessity for now.
     */
    if (this.nodesMarkedForDeath.length == 0) {
      setTimeout(abacusKillMarkedNodes, 0);
    }
    this.nodesMarkedForDeath.push(aNode);
    return true;
  },

  killMarkedNodes: function() {
    while (this.nodesMarkedForDeath.length > 0) {
      var aNode = this.nodesMarkedForDeath.shift();
      aNode.parentNode.removeChild(aNode);
    }
    return true;
  },

  updateCommandsDisabled: function(aNode) {
    if (!this.hasInitialized) {
      return;
    }
    for (var i = 0; i < this.commandsArray.length; i++) {
      var aCommand = this.commandsArray[i];
      var disabled = (eval(aCommand.id)).isCommandDisabled(aNode);
      if (aCommand.id == "abacusOk") {
        aCommand = this.acceptButton;
      }
      if (disabled) {
        aCommand.setAttribute("disabled", true);
      } else {
        aCommand.removeAttribute("disabled");
      }
    }
  },

  toString: function() {
    return "[object abacusTemplateDialogEditor]";
  }
}

function updateEditorTemplate() {
  editor.updateTemplate();
}

function abacusKillMarkedNodes() {
  editor.killMarkedNodes();
}

const abacusUpdateTemplate = {
  isCommandDisabled: function() {
    return !editor.selectedTemplate;
  }
};

const abacusOk = {
  isCommandDisabled: function() {
    return !testTemplate();
  }
}

function watchUserErrors(methodName) {
  assert(userErrorBox, "There must be a user error box in place!", true);
  var rv = false;
  var args = [];
  for (var i = 1; i < arguments.length; i++) {
    args[i - 1] = arguments[i];
  }
  userErrorBox.value = "";
  try {
   if (ecmaDebug) {
      args.unshift(this);
      args.unshift(this[methodName]);
      rv = applyContract.apply(window, args); // we might as well enforce contracts when available
    } else {
      rv = this[methodName].apply(this, args);
    }
  }
  catch(e) {
    if (e.name == "UserError") {
      userErrorBox.value = e.message;
    } else {
      throw e;
    }
  }
  return rv;
}

function doAccept() {
  switch (aTemplate.templateEditCode) {
    case 0: 
      aTemplate.templateEditCode = 1;
    case 1:
    case 3:
      break;
    case 4:
      aTemplate.templateEditCode = 3;
      break;
    default:
      throw new Error("An unauthorized templateEditCode (" + aTemplate.templateEditCode + ") at doAccept!");
  }
  var markupGood = testTemplate();
  if (markupGood) {
    aTemplate.templateSource = mainText.value;
    opener.editor.watchUserErrors('templateUpdated', aTemplate);
    return true;
  }
  return false;    
}

function doCancel() {
  opener.editor.watchUserErrors('templateCanceled', aTemplate);
  return true;
}

ecmaDebug = opener.ecmaDebug;
const parser = opener.parser;
const aTemplate = arguments[0];
assert(aTemplate.templateEditCode != 2, "Template edit code 2 is prohibited in templateDialog.xul!", true);
assert(aTemplate.templateEditCode != 5, "Template edit code 5 is prohibited in templateDialog.xul!", true);

var prefixText;
var mainText;
var postfixText;
var userErrorBox;
window.title = aTemplate.dialogTitle;

window.addEventListener("load", function() {
  userErrorBox = document.getElementById("usererrorbox");
  prefixText = document.getElementById("prefixText");
  var basePrefix = aTemplate.baseTemplate.basePrefix;
  prefixText.value = basePrefix;
  mainText = document.getElementById("mainText");
  mainText.value = aTemplate.templateSource;
  postfixText = document.getElementById("postfixText");
  var re= /[\s\>]/;
  var tagNameEnds = basePrefix.search(re);
  var rootElement = basePrefix.substring(1, tagNameEnds);
  postfixText.value = "</" + rootElement +">";
  editor.init();
}, true);
