/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Abacus MathML Root Node Preparer.
 *
 * The Initial Developer of the Original Code is 
 * Alexander J. Vincent <ajvincent@juno.com>.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/***** WARNING: LARGE PORTIONS OF THIS CODE ARE UNDEVELOPED AND/OR UNTESTED!  USE AT YOUR OWN RISK!

  Methods to implement list:
  * mathNodePreparer.parsePresentationMarkup:  Tries to return a content MathML node and a corresponding presentation 
                                               MathML node, given a presentation MathML node as its first argument.
  Necessary to support:
  <math><mrow><foo/><bar/></mrow></math>

  * mathNodePreparer.parseContentMarkup:       Tries to return a content MathML node and a corresponding presentation 
                                               MathML node, given a content MathML node as its first argument.
  Necessary to support:
  <math><apply><foo/><bar/></apply></math>

  * mathNodePreparer.matchElementToPattern:    Reserved name for function designed to support parseContentMarkup and
                                               parsePresentationMarkup.  This function may or may not be used.

  Methods needing additional support work:
  * getMathElementType                         Needs a more comprehensive list of content and presentation MathML elements.

  * prepareMathNode                            Near the end, needs a more complete specification for conditions under which
                                               it should call this.insertMEditNode().

  The entire script is very large, and there's probably plenty of room for optimization.  Suggestions are welcome.

 *****/

const mathNodePreparer = {
  // NodeFilters

  acceptNodeFilter: {
    acceptNode: function(aNode) {
      return NodeFilter.FILTER_ACCEPT;
    },

    toString: function() {
      return "[object abacusNodeFilter]";
    }
  },

  hasMEditIndexFilter: {
    requiredAncestorNode: null,

    acceptNode: function(aNode) {
      var index = aNode.getAttributeNS(MEDIT_NS, "index");
      if ((index == "")||(isNaN(index))) {
        return NodeFilter.FILTER_SKIP;
      }
      index *= 1;
      if ((Math.floor(index) != index) || (index < 0)) {
        return NodeFilter.FILTER_SKIP;
      }
      if (this.requiredAncestorNode) {
        var bNode = aNode;
        while ((bNode)&&(bNode != this.requiredAncestorNode)) {
          bNode = bNode.parentNode;
        }
        if (!bNode) {
          return NodeFilter.FILTER_SKIP;     
        }
      }
 
      return NodeFilter.FILTER_ACCEPT;
    },

    toString: function() {
      return "[object abacusNodeFilter]";
    }
  },

  hasIdOrXrefAttributeFilter: {
    acceptNode: function(aNode) {
      if (aNode.hasAttribute("id")||(aNode.hasAttribute("xref"))) {
        /* It's dumb to assume someone's going to prefix their ID attributes.
           Though it's probably almost as dumb to assume someone won't prefix their ID attributes, there's very
           little reason to prefix one.

           As for xref attributes... though I'd prefer that we prefixed them with math: on non-MathML elements,
           XHTML 1.1 + MathML 2.0 DTD doesn't do that naturally.  This is probably because of the Modularization of
           XHTML Recommendation from the W3C, and I can see why they do it this way.  It's for the same reason I prefix
           every mEdit:index node: uniformity.
        */
        return NodeFilter.FILTER_ACCEPT;
      }
      return NodeFilter.FILTER_SKIP;
    },

    toString: function() {
      return "[object abacusNodeFilter]";
    }
  },

  // convenience function to bypass comments, processing instructions, text nodes, etc. in child nodes

  getChildElements: function(rootNode) {
    var walker = document.createTreeWalker(rootNode, NodeFilter.SHOW_ELEMENT, this.acceptNodeFilter, true);
    var output = "";
    var rv = [];
    if (walker.firstChild()) {
      rv.push(walker.currentNode)
      while (walker.nextSibling()) {
        rv.push(walker.currentNode);
      }
    }
    return rv;
  },

// functions to parse markup

  matchElementToPattern: function(aElement) {
    throw new UserError("mathNodePreparer.matchElementToPattern() is not implemented yet!");
  },

  parsePresentationMarkup: function(aNode) {
    throw new UserError("math:presentation translation of mathChildElement not implemented yet!");
    warn("Pure presentation markup makes for a content markup guessing game, per the MathML 2.0 Recommendation.");
    var rv = {
      contentNode: null,
      presentationNode: aNode
    };
    return rv;
  },

  parseContentMarkup: function(aElement) {
    throw new UserError("math:content translation of mathChildElement not implemented yet!");
  },

// identify element as presentation MathML, content MathML, or a different markup language

  getMathElementType: function(aElement) {
    var rv;
    var errMsg = "I don't know what math element type to call " + aElement.nodeName;
    switch (aElement.namespaceURI) {
      case MATH_NS:
        rv = "math:";
        break;
        // <math:foo/> elements are the only ones we have to go to the next step to identify.
        // All other namespaces should directly return a simple flag identifying the namespace.

      case XHTML_NS:
        return "html";

      default:
        throw new UserError(errMsg);
    }

    switch (aElement.localName) {
      // special cases!
      case "semantics":
      case "annotation-xml":
      case "annotation":
        rv += aElement.localName;
        break;

      case "apply":
      case "ci":
      case "cn":
        rv += "content";
        break;

      case "mrow":
      case "mi":
      case "mn":
        rv += "presentation";
        break;

      default:
        throw new UserError(errMsg);
    }

    return rv;
  },

  // report if several nodes have the same type

  matchMathElementTypes: function(aElementArray) {
    var matchBase;
    matchBase = this.getMathElementType(aElementArray[0]);
    for (var i = 1; i < aElementArray.length; i++) {
      if (matchBase != this.getMathElementType(aElementArray[i])) {
        return false;
      }
    }
    return matchBase;
  },

  // identify mEdit:index values for later setting of attributes

  addMEditIndexValues: function(rootElement, rv) {
    assert(rv instanceof Array, "We need an array to return from addMEditIndexValues!", true);
    var index;
    var errMsg;
    var indexNodes = [];
    var j = -1;
    var walker = document.createTreeWalker(rootElement, NodeFilter.SHOW_ELEMENT, this.hasMEditIndexFilter, true);
    if (this.hasMEditIndexFilter.acceptNode(rootElement) == NodeFilter.FILTER_ACCEPT) {
      index = walker.currentNode.getAttributeNS(MEDIT_NS, "index") * 1;
      errMsg = "The root node of a semantics branch should have mEdit:index='0' or no mEdit:index at all!";
      assertAboutInput(index === 0, errMsg, true);
      indexNodes[0] = rootElement;
      walker.currentNode.mEditIndex = 0; // this is only temporary!
      j = 0;
    }

    while (walker.nextNode()) {
      index = walker.currentNode.getAttributeNS(MEDIT_NS, "index") * 1;
      errMsg = "Duplicate mEdit:index value detected in a <semantics> branch!"
      assertAboutInput(typeof indexNodes[index] == "undefined", errMsg, true);
      assertAboutInput(index != 0, "Only the root node of a semantics branch may have mEdit:index='0'!", true);
      indexNodes[index] = walker.currentNode;
      walker.currentNode.mEditIndex = index; // this is only temporary!
      j = Math.max(j, index);
    }

    errMsg = "Missing mEdit:index value detected in a <semantics> branch!";
    assertAboutInput(j == indexNodes.length - 1, errMsg, true);
    for (i = 0; i < indexNodes.length; i++) {
      assertAboutInput(typeof indexNodes[index] != "undefined", errMsg, true);
    }

    if (rv.length != 0) {
      assertAboutInput(j == rv.length - 1, errMsg, true);
      for (i = 0; i < rv.length; i++) {
        assertAboutInput(typeof rv[index] != "undefined", errMsg, true);
      }
    }

    for (var i = 0; i < indexNodes.length; i++) {
      if (typeof rv[i] == "undefined") {
        rv[i] = [];
      }
      rv[i].push(indexNodes[i]);
    }
  },

  addIdAndXrefValues: function(rootElement, rv) {
    assert(!(rv instanceof Array), "rv must not be an array in this.addIdAndXRefValues!", true);
    var indexNodes = {length: 0};
    var ns;
    var xref;
    var errMsg;
    var i;
    var j;
    var k;
    var walker = document.createTreeWalker(rootElement, NodeFilter.SHOW_ELEMENT, this.hasIdOrXrefAttributeFilter, true);
    if (this.hasIdOrXrefAttributeFilter.acceptNode(rootElement) == NodeFilter.FILTER_ACCEPT) {
      xref = rootElement.getAttribute("id")||rootElement.getAttribute("xref");
      assert(xref, "Failed to get id or xref attribute from " + rootElement.nodeName, true);
      errMsg = "<foo id='length'/> is illegal in expressions passed to Abacus.  (Sorry.)";
      assertAboutInput(xref != "length", errMsg, true);
      indexNodes[xref] = rootElement;
      indexNodes.length++;
    }

    while (walker.nextNode()) {
      xref = walker.currentNode.getAttribute("id")||walker.currentNode.getAttribute("xref");
      assert(xref, "Failed to get id or xref attribute from " + walker.currentNode.nodeName, true);
      errMsg = "Duplicate id/xref value detected in a <semantics> branch!"
      assertAboutInput(typeof indexNodes[xref] == "undefined", errMsg, true);
      errMsg = "<foo id='length'/> is illegal in expressions passed to Abacus.  (Sorry.)";
      assertAboutInput(xref != "length", errMsg, true);
      indexNodes[xref] = walker.currentNode;
      indexNodes.length++;
    }

    if ((typeof rv.length == "number")&&(rv.length != 0)) {
      errMsg = "Previously detected " + rv.length + " nodes, now found " + indexNodes.length;
      assertAboutInput(indexNodes.length == rv.length, errMsg, true);
      for (i in rv) {
        if (i == "length") {
          continue;
        }
        errMsg = "Expected an array at rv[" + i + "]";
        assert(typeof rv[i] != "undefined", errMsg, true);
        assert(rv[i] instanceof Array, errMsg, true);
      }
    }

    for (i in indexNodes) {
      if (i == "length") {
        rv.length = indexNodes.length;
        continue;
      }
      if (typeof rv[i] == "undefined") {
        rv[i] = [];
      }
      rv[i].push(indexNodes[i]);
    }

    return rv;
  },

  // replace an element with a <mEdit:fitb mEdit:index="n"/> element

  insertMEditNode: function(indexNode, index) {
    var mEditNode = document.createElementNS(MEDIT_NS, "mEdit:fitb");
    mEditNode.setAttributeNS(MEDIT_NS, "mEdit:index", index);
    mEditNode.setAttribute("value", indexNode.firstChild.nodeValue);
    indexNode.parentNode.replaceChild(mEditNode, indexNode);
  },

  // The critical function.

  prepareMathNode: function(aNode) {
    /* We do a lot of dancing around comments, processing instructions, etc. because Abacus MathML Editor only cares 
     * about elements.  It does not provide a capability for editing, removing, or inserting non-element nodes.
     * Abacus MathML Editor will move non-element nodes to keep contextual information in place.
     * Abacus MathML Editor will also remove whitespace text nodes as necessary.
     *
     * In a sense, this function does some basic MathML validation.  It does not substitute for a validating XML parser.
     * The partial validation this function controls is just enough to assure that Abacus MathML Editor will correctly 
     * handle a <math:math> node passed to it.
     *
     * You'll see a lot of calls to assertAboutInput() in this function, as opposed to assert().  This is because we are
     * directly working with a value the user gave us (a MathML element) and its contents.  For the same reason, you'll
     * also see some calls to "throw new UserError()".
     */

    var walker;
    var childElementsArray;
    var grandChildElementsArray;
    var mathNodeSet;
    var errMsg;
    var mathChildElement;
    var rowType;
    var encoding;
    var semanticsNode;
    var semanticsMainNode;
    var mEditNode;
    var content;
    var contentDocFragment;
    var annotationDocFragment;
    var annotationNode;
    var contentFound;
    var presentationNode;
    var index;
    var indexNode;
    var indexArray;
    var indexNodeArray;
    var contentIndexes;
    var idPrefix;
    var i;
    var j;
    var k;

    removeWhitespaceByNamespace(aNode, MATH_NS);
    removeWhitespaceByNamespace(aNode, MEDIT_NS);
    errMsg = "Character data nodes are prohibited as direct children of the math:math element!";
    for (i = 0; i < aNode.childNodes.length; i++) {
      var child = aNode.childNodes[i];
      // "Magic numbers" are evil, I know...
      assertAboutInput((child.nodeType != 3)&&(child.nodeType != 4), errMsg, true);
    }

    childElementsArray = this.getChildElements(aNode);

    switch (childElementsArray.length) {
      case 0:
        // We were passed a new math node.  Set it up and bail out.
        semanticsNode = document.createElementNS(MATH_NS, "semantics");
        semanticsNode.setAttribute("encoding", "MathML-Content");
        mEditNode = document.createElementNS(MEDIT_NS, "mEdit:fitb");
        mEditNode.setAttributeNS(MEDIT_NS, "mEdit:index", 0);
        semanticsNode.appendChild(mEditNode);

        annotationNode = document.createElementNS(MATH_NS, "annotation-xml");
        annotationNode.setAttribute("encoding", "Abacus-Template-Info");
        annotationNode.appendChild(mEditNode.cloneNode(true));
        semanticsNode.appendChild(annotationNode);

/*
        annotationNode = document.createElementNS(MATH_NS, "annotation-xml");
        annotationNode.setAttribute("encoding", "MathML-Presentation");
        // Show me a MathML branch
        annotationNode.setAttributeNS(MEDIT_NS, "mEdit:visibleBranch", "true");
        annotationNode.appendChild(mEditNode.cloneNode(true));
        semanticsNode.appendChild(annotationNode);
*/

        aNode.appendChild(semanticsNode);
        aNode.setAttributeNS(MEDIT_NS, "mEdit:editorURL", "chrome://abacus/content/math/editor.xul");
        aNode.setAttributeNS(XMLNS_NS, "xmlns:mEdit", MEDIT_NS);
        return true;
      
      case 1:
        mathChildElement = childElementsArray[0];
        rowType = this.getMathElementType(mathChildElement);
        errMsg = "<math:math><" + mathChildElement.nodeName + "/></math:math> is illegal!";
        assertAboutInput(rowType.substr(0, 5) == "math:", errMsg, true);
        break;

      default:
        /* At least two elements are direct children of the <math:math> element.  This is not good, not at all!
         * If we get past this part, we'll have a single <math:mrow> element.  
         * Presentation markup, however, is the hardest to figure out.
         */
        errMsg = "All children of a math element branch must be of the same content type!";
        rowType = this.matchMathElementTypes(childElementsArray);
        assertAboutInput(rowType, errMsg, true);
        errMsg = "Only presentation MathML elements may be grouped as direct children of a <math:math/> element!";
        assertAboutInput(rowType == "math:presentation", errMsg, true);
        mathChildElement = document.createElementNS(MATH_NS, "math:mrow");
        while (aNode.firstChild) {
          mathChildElement.appendChild(mathChildElement);
        }
        aNode.appendChild(mathChildElement);
    }

    // Okay, mathChildElement is our one and only element which is a child node of <math:math/>.
    // mathChildElement is also a node in the MathML namespace.

    switch (rowType) {
      case "math:presentation":
        // Worst-case scenario.  We have to infer content markup from presentation markup.
        mathNodeSet = this.parsePresentationMarkup(presentationNode);
        semanticsNode = document.createElementNS(MATH_NS, "semantics");
        semanticsNode.setAttribute("encoding", "MathML-Content");
        semanticsNode.appendChild(mathNodeSet.contentNode);

        annotationNode = document.createElementNS(MATH_NS, "annotation-xml");
        annotationNode.setAttribute("encoding", "MathML-Presentation");
        annotationNode.appendChild(mathNodeSet.presentationNode);
        aNode.appendChild(semanticsNode);
        break;

      case "math:content":
        // This is not as bad.  We only have to guess at the presentation markup, and content markup is pretty exact.
        mathNodeSet = this.parseContentMarkup(mathChildElement);
        semanticsNode = document.createElementNS(MATH_NS, "semantics");
        semanticsNode.setAttribute("encoding", "MathML-Content");
        semanticsNode.appendChild(mathNodeSet.contentNode);

        annotationNode = document.createElementNS(MATH_NS, "annotation-xml");
        annotationNode.setAttribute("encoding", "MathML-Presentation");
        annotationNode.appendChild(mathNodeSet.presentationNode);
        semanticsNode.appendChild(annotationNode);
        aNode.appendChild(semanticsNode);
        
        break;

      case "math:semantics":
        // Fantastic!  We should have lots of id and xref attributes to keep us happy.  Or even better, mEdit:index.
        // Let's find our content MathML fragment.
        semanticsNode = mathChildElement;
        encoding = semanticsNode.getAttribute("encoding");
        if (encoding == "MathML-Content") {
          // Better and better!
          annotationNode = this.getChildElements(mathChildElement)[0];
          errMsg = "<semantics encoding='MathML-Content'><" + annotationNode.nodeName + "/></semantics>: illegal";
          assertAboutInput(annotationNode.namespaceURI == MATH_NS, errMsg, true);
          assertAboutInput(this.getMathElementType(annotationNode) == "math:content", errMsg, true);
          break;
        }

        j = null;
        contentFound = false;
        presentationNode = null;
        childElementsArray = this.getChildElements(mathChildElement);
        semanticsMainNode = childElementsArray[0];
        if (encoding == "MathML-Presentation") {
          errMsg = "<semantics encoding='MathML-Presentation'><" + semanticsMainNode.nodeName + "/></semantics>: illegal";
          assertAboutInput(semanticsMainNode.namespaceURI == MATH_NS, errMsg, true);
          assertAboutInput(this.getMathElementType(semanticsMainNode) == "math:presentation", errMsg, true);
          presentationNode = semanticsMainNode;
        } else {
          errMsg = "<semantics encoding='" + encoding + "'>" + semanticsMainNode.nodeName + "/></semantics: illegal";
          warn("We can't check everything.  For example, <semantics encoding='XHTML'><svg:svg/></semantics> passes.");
          assertAboutInput(semanticsMainNode.namespaceURI != MATH_NS, errMsg, true);
        }

        for (i = 1; i < childElementsArray.length; i++) {
          annotationNode = childElementsArray[i];
          // if we've gotten here, then we know about what the first child of <semantics/> is.
          errMsg = "Illegal annotationNode: " + annotationNode.nodeName;
          assertAboutInput(annotationNode.namespaceURI == MATH_NS, errMsg, true);
          if (annotationNode.getAttribute("encoding") == "MathML-Content") {
            // Good, but we need to switch things around here.
            semanticsNode.setAttribute("encoding", "MathML-Content");
            annotationNode.setAttribute("encoding", encoding);
            contentDocFragment = document.createDocumentFragment();
            annotationDocFragment = document.createDocumentFragment();
            while (annotationNode.firstChild) {
              contentDocFragment.appendChild(annotationNode.firstChild);
            }
            while (childElementsArray[1].previousSibling) {
              annotationDocFragment.appendChild(mathChildElement.firstChild);
            }
            assert(semanticsNode.firstChild == childElementsArray[1], "We've made a mistake somewhere!", true);
            semanticsNode.insertBefore(contentDocFragment, semanticsNode.firstChild);
            annotationNode.appendChild(annotationDocFragment);
            contentFound = true;
            break; // the switch is complete!
          }

          if ((annotationNode.getAttribute("encoding") == "MathML-Presentation")&&(!presentationNode)) {
            presentationNode = annotationNode;
          }
        }

        if (!contentFound) {
          // Houston, we have a problem.
          if (!presentationNode) {
            // A BIG PROBLEM.
            throw new UserError("Abacus MathML editor needs at least one MathML markup branch!");
            /* In theory, this is not necessarily true.  In theory, if Abacus has the templates available to search, 
               and the right algorithms, we can build both a content MathML branch of the semantics element, 
               and a presentation MathML branch as well.  

               If you want to implement this (perhaps using other code in this script as an inspiration), be my guest.

               Seriously, anyone who writes a MathML <semantics/> element without including at least one MathML markup 
               branch is certifiable, and probably certified already.  Either as a genius, an engineer, or a lunatic.
             */
          }
          /* here, we have a presentationNode, but no contentNode.
<math:semantics encoding="foo">
  <j:foo/>
  <math:annotation-xml encoding="bar"/>
  <math:annotation-xml encoding="MathML-Presentation"/>
  <math:annotation encoding="baz"/>
  <!-- etc., etc. etc. -->
</math:semantics>
           */
          mathNodeSet = this.parsePresentationMarkup(presentationNode);
          semanticsNode.setAttribute("encoding", "MathML-Content");
          annotationNode = document.createElementNS(MATH_NS, "annotation-xml");
          annotationNode.setAttribute("encoding", encoding);
          annotationDocFragment = document.createDocumentFragment();
          childElementsArray = this.getChildElements(semanticsNode);
          while (childElementsArray[1].previousSibling) {
            annotationDocFragment.appendChild(mathChildElement.firstChild);
          }
          semanticsNode.insertBefore(mathNodeSet.contentNode, semanticsNode.firstChild);
          annotationNode.appendChild(annotationDocFragment);
          semanticsNode.appendChild(annotationNode);
          contentFound = true;
        }
        break; // end select (rowType) {case: "math:semantics"};

      case "math:annotation":
      case "math:annotation-xml":
        // Sigh.
        throw new UserError("<math:math><annotation/></math:math>, <math:math><annotation-xml/></math:math> are illegal!");
        break;

      default:
        assert(false, "Expected a node in the MathML namespace, received " + mathChildElement, true);
    } // end switch (rowType)

    /* If we've reached this point, we've cleared some hurdles.  Our markup now should look like:
<math:math>
  <math:semantics encoding="MathML-Content">
    <math:apply/><!-- Probably -->
    <math:annotation-xml encoding="foo"/>
    <math:annotation-xml encoding="MathML-Presentation">
      <math:mrow/><!-- Probably -->
    </math:annotation-xml>
    <math:annotation encoding="bar"/>
    <math:annotation-xml encoding="baz"/>
    <!-- etc., etc., etc. -->
  </math:semantics>
</math:math>

    It's now time to start parsing this expression for id and xref attributes, which will then give us our mEdit:index
    values.  We first look for mEdit:index values already resident in the expression.

    Though it should look like this, a few assert() calls to verify what we see don't hurt.  If we fail an assert() call
    here, then there's probably a bug earlier in the code checking algorithms.  In other words, if you're hacking in this
    section, do not assume all values before this section can be trusted.  (This is a pretty big function, after all...)
    */

    errMsg = "Expected a semantics element, received " + semanticsNode;
    assert(semanticsNode, errMsg, true);
    errMsg = "Expected <math:semantics/>, received " + semanticsNode.nodeName;
    assert(semanticsNode.namespaceURI == MATH_NS, errMsg, true);
    assert(semanticsNode.localName == "semantics", errMsg, true);

    childElementsArray = this.getChildElements(semanticsNode);
    assert(childElementsArray.length > 1, "Expected at least two child elements of <math:semantics/>", true);
    errMsg = "Expected MathML content node, received <" + childElementsArray[0].nodeName + "/>";
    assert(this.getMathElementType(childElementsArray[0]) == "math:content", errMsg, true);

    indexArray = [];
    this.addMEditIndexValues(childElementsArray[0], indexArray);
    contentIndexes = indexArray.length + 0;

    for (i = 1; i < childElementsArray.length; i++) {
      errMsg = "Expected <math:annotation/> or <math:annotation-xml/> node, received <" + childElementsArray[i] + "/>";
      assert(childElementsArray[i].namespaceURI == MATH_NS, errMsg, true);
      if (childElementsArray[i].localName == "annotation") {
        continue; // we don't want to bother trying to get mEdit:index values for non-XML annotations
      }
      assert(childElementsArray[i].localName == "annotation-xml", errMsg, true);
      grandChildElementsArray = this.getChildElements(childElementsArray[i]);
      errMsg = "Abacus MathML Editor requires each branch have exactly one root node!";
      assertAboutInput(grandChildElementsArray.length == 1, errMsg, true);
      /* Though MathML may explicitly permit the following:

         <math:annotation-xml encoding="XHTML">
           <html:p>Hello World</html:p>
           <html:p>This is a test</html:p>
         </math:annotation-xml>

         Abacus MathML Editor cannot handle that.  It would be safer to say:

         <math:annotation-xml encoding="XHTML">
           <html:div>
             <html:p>Hello World</html:p>
             <html:p>This is a test</html:p>
           </html:div>
         </math:annotation-xml>

         The reason Abacus MathML Editor requires that is it must assign a mEdit:index="0" attribute to a root node 
         of each branch.  It thus needs to be assured there is precisely one such root node.
       */
      this.addMEditIndexValues(grandChildElementsArray[0], indexArray);
    }
    errMsg = "A non-content branch of the semantics element had mEdit:index values the content branch didn't have!";
    assertAboutInput(contentIndexes == indexArray.length, errMsg, true);

    if (contentIndexes == 0) {
      /* No problem, we somewhat expect this.  Go for the id and xref attributes.
         In fact, if we don't hit this code, then the application offering us the MathML node is mEdit-aware and passed
         us a friendly expression.
       */
      indexArray = {};
      this.addIdAndXrefValues(childElementsArray[0], indexArray);
      contentIndexes = indexArray.length + 0; // once again
      for (i = 1; i < childElementsArray.length; i++) {
        errMsg = "Expected <math:annotation/> or <math:annotation-xml/> node,";
        errMsg+= " received <" + childElementsArray[i].nodeName + "/>";
        if (childElementsArray[i].localName == "annotation") {
          continue; // we don't want to bother trying to get mEdit:index values for non-XML annotations
        }
        grandChildElementsArray = this.getChildElements(childElementsArray[i]);
        this.addIdAndXrefValues(grandChildElementsArray[0], indexArray);
      }
      errMsg = "A non-content branch of the semantics element had mEdit:index values the content branch didn't have!";
      assertAboutInput(contentIndexes == indexArray.length, errMsg, true);
      assert(contentIndexes > 0, "Expected at least one content index", true);

      // It sure would be nice if every id/xref we got was equal to the mathNode's id plus a colon plus a unique number.
      // Abacus outputs MathML markup in that fashion.

      i = true;
      idPrefix = aNode.getAttribute("id") + ":";
      k = [];
      for (j in indexArray) {
        if (j == "length") {
          continue;
        }
        if (j.substr(0, idPrefix.length) != idPrefix) {
          i = false;
          break;
        }
        index = j.substr(idPrefix.length) * 1;
        if ((isNaN(index))||(index < 0)||(index != Math.floor(index))) {
          i = false;
          break;
        }
        k.length = Math.max(k.length, index + 1);
        k[index] = indexArray[j];
      }

      if (i) {
        // Let's be sure, though, that it's properly formatted.
        for (j = 0; j < k.length; j++) {
          i = assert(typeof k[j] != "undefined", "We're missing a urn:abacus:foo:" + j + " value.");
          // Though this is an error from the application passing us the node, we can recover from it.
          // This is an assert() and not an assertAboutInput() for that reason.
          if (!i) {
            k.length = 0; // truncate and bail out at the same time
          }
        }
      }

      if (!i) {
        // Assign new mEdit:index values.
        i = 0;
        for (j in indexArray) {
          if (j == "length") {
            continue;
          }
          k[i] = indexArray[j];
          i++;
        }
      }
      indexArray = k;
    } // end if (contentIndexes == 0)

    /* Last step:  assign mEdit:index attributes to each node as needed
       If we've reached this point, then the MathML node passed to Abacus MathML Editor is very stable.
       We will still need to check, after assigning the attributes, that each semantics branch's root node has an
       mEdit:index value of 0.  But beyond that, we're done.
    */

    assert(indexArray.length > 0, "Failed to retrieve any mEdit:index values!", true);
    for (i = 0; i < indexArray.length; i++) {
      indexNodeArray = indexArray[i];
      assert(indexNodeArray, "Couldn't get indexArray[" + i + "]", true);
      for (j = 0; j < indexNodeArray.length; j++) {
        indexNode = indexNodeArray[j];

        switch (indexNode.namespaceURI) {
          case MATH_NS:
            switch (indexNode.localName) {
              case "cn":
              case "ci":
              case "mn":
              case "mi":
                this.insertMEditNode(indexNode, i);
                break;
            } // end switch (indexNode.localName) for case MATH_NS:
        } // end switch (indexNode.namespaceURI);

        if (indexNode.parentNode) {
          indexNode.setAttributeNS(MEDIT_NS, "mEdit:index", i);
        }
      }
    }

    // Final checks
    childElementsArray = this.getChildElements(semanticsNode);
    errMsg = "Failed to set mEdit:index='0' on a semantics branch root node!";
    assert(childElementsArray[0].getAttributeNS(MEDIT_NS, "index") == "0", errMsg, true);
    j = true;
    for (i = 1; i < childElementsArray.length; i++) {
      // We already did assertions to ensure all child elements of <math:semantics/> are either <math:annotation/>
      // or <math:annotation-xml/>
      if (childElementsArray[i].localName == "annotation") {
        continue;
      }
      annotationNode = childElementsArray[i];
      if (j) {
        // Show me a MathML branch
        annotationNode.setAttributeNS(MEDIT_NS, "mEdit:visibleBranch", "true");
        j = false;
      }
      grandChildElementsArray = this.getChildElements(annotationNode);
      // We already did assertions to ensure annotationNode has exactly one child element.
      assert(grandChildElementsArray[0].getAttributeNS(MEDIT_NS, "index") == "0", errMsg, true);
    }
    childElementsArray = this.getChildElements(aNode)
    errMsg = "Expected exactly one child element of aNode, received " + childElementsArray.length;
    assert(childElementsArray.length == 1, errMsg, true);
    errMsg = "Expected <math:semantics/> as only child of <math:math/>, found " + childElementsArray[0].nodeName;
    assert(childElementsArray[0] == semanticsNode, errMsg, true);

    // Show me a MathML branch
    return true;
  },

  toString: function() {
    return "[object mathNodeLoader]";
  }
}