/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Abacus Transaction Constructors.
 *
 * The Initial Developer of the Original Code is 
 * Alexander J. Vincent <ajvincent@juno.com>.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

function nsITransaction() {
}
nsITransaction.prototype = {
  doTransaction: function() {
  },

  undoTransaction: function() {
  },

  redoTransaction: function() {
    this.doTransaction();
  },

  isTransient: false,

  merge: function(aTransaction) {
    return false;
  },

  toString: function() {
    return "[object nsITransaction]";
  }
}

const nsITransactionManager = Components.interfaces.nsITransactionManager;

function abacusTransaction(txmgr, mathNode, originalContents, idPrefix) {
  this.txmgr = txmgr;
  this.mathNode = mathNode;
  this.originalContents = originalContents;
  this.parentNode = typeof originalContents.parentNode != "undefined" ? originalContents.parentNode : arguments[4];
  this.idPrefix = idPrefix;
  this.backupCopy = null;
}
abacusTransaction.prototype = new nsITransaction();
abacusTransaction.prototype.doTransaction = function() {
  if (this.originalContents) {
    if (this.originalContents instanceof Components.interfaces.nsIDOMRange) {
      this.backupCopy = this.originalContents.replaceContents(this.mathNode);
    } else {
      this.parentNode.replaceChild(this.mathNode, this.originalContents);
    }
  } else {
    this.parentNode.appendChild(this.mathNode);
  }

  if (typeof this.localAppDoTxn == "function") {
    this.localAppDoTxn();
  }
}
abacusTransaction.prototype.undoTransaction = function() {
  if (typeof this.localAppUndoTxn == "function") {
    this.localAppUndoTxn();
  }

  if (this.originalContents) {
    if (this.backupCopy) {
      this.parentNode.replaceChild(this.backupCopy, this.mathNode)
    } else {
      this.parentNode.replaceChild(this.originalContents, this.mathNode);
    }
  } else {
    this.parentNode.removeChild(this.parentNode);
  }
}
abacusTransaction.prototype.redoTransaction = function() {
  this.doTransaction();
}
abacusTransaction.prototype.openDialog = function(aRuntimeConfig) {
  window.openDialog("chrome://abacus/content/math/editor.xul", "Abacus MathML Editor",
                    "chrome, modal=1, width=640, height=480", this, aRuntimeConfig);
}
abacusTransaction.prototype.acceptMethods = function(aObject) {
  for (var prop in aObject) {
    if (typeof aObject[prop] == "function") {
      this[prop] = aObject[prop];
    }
  }
}
